#include <iostream>
#include <time.h>
#include <stdlib.h>
#include "Comm.h"

using namespace std;

#define DEBUG 0
#define MAXLENGTH 1000

Comm *Comm::instance = NULL;
char Comm::ONE = 1;
char Comm::ZERO = 0;

Comm::Comm(int argc, char *argv[]) {
  MPI_Init(&argc, &argv);
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  MPI_Comm_size(MPI_COMM_WORLD, &processes);

  state = WAITING;
  inquiry = false;
  requestActive = false;
  terminated = false;
  requestTotal = 0;
  requestSucc = 0;

  cout << "Proces " << rank << ": Inicializace komunikace" << endl;
  srand(time(NULL));
  donor = rand() % processes;
  if(donor == rank) donor = (donor + 1) % processes;
  
  char *name = new char [10];
  sprintf(name, "out_p%d", rank);
  fout = freopen(name, "a", stdout);

  cout << endl << "-----------------------------" << endl;
  cout << "Pocet procesu: " << processes << endl;
  cout << "-----------------------------" << endl;
  cout << "Proces " << rank <<  ": Konec inicializace komunikace" << endl;
  cout << flush;
}

void Comm::sendGraph(Graph *graph) {
  MPI_Request request;
  int nodes = graph->getNodesCount();
  const int length = nodes * nodes * sizeof(char) + sizeof(int);

  if(length > MAXLENGTH) {
    cout << "Proces " << rank << ": ";
    cout << "Maximalni povolena delka zpravy je " << MAXLENGTH << "B (pozadovano " << length << "B)" << endl;
    cout << flush;

    exit(1);
  }

  char buffer[length];
  int position = 0;
  bool **g = graph->getGraph();

  MPI_Pack(&nodes, 1, MPI_INT, buffer, length, &position, MPI_COMM_WORLD);

  for(int i = 0; i < nodes; i++)
    for(int j = 0; j < nodes; j++)
      if(g[i][j]) MPI_Pack(&ONE, 1, MPI_CHAR, buffer, length, &position, MPI_COMM_WORLD);
      else MPI_Pack(&ZERO, 1, MPI_CHAR, buffer, length, &position, MPI_COMM_WORLD);

  for(int i = 0; i < nodes; i++) delete [] g[i];
  delete [] g;
     
  for(int i = 1; i < processes; i++)
    MPI_Isend(buffer, position, MPI_PACKED, i, GRAPH, MPI_COMM_WORLD, &request);

  cout << "Proces " << rank << ": ";
  cout << "Vsem procesum rozeslana matice sousednosti o " << nodes << " uzlech (" << position << "B)" << endl;
  cout << flush;
}

Graph *Comm::receiveGraph() {
  const int length = MAXLENGTH;
  char buffer[length];
  int position = 0;

  MPI_Recv(buffer, length, MPI_PACKED, MPI_ANY_SOURCE, GRAPH, MPI_COMM_WORLD, &status);

  int nodes = 0;
  MPI_Unpack(buffer, length, &position, &nodes, 1, MPI_INT, MPI_COMM_WORLD);

  bool **g = new bool *[nodes];
  for(int i = 0; i < nodes; i++) g[i] = new bool [nodes];

  for(int i = 0; i < nodes; i++)
    for(int j = 0; j < nodes; j++)
      MPI_Unpack(buffer, length, &position, &g[i][j], 1, MPI_CHAR, MPI_COMM_WORLD);

  Graph *graph = new Graph(nodes, g);

  for(int i = 0; i < nodes; i++) delete [] g[i];
  delete [] g;
  
  cout << "Proces " << rank << ": ";
  cout << "Prijata matice sousednosti o " << nodes << " uzlech (" << position << "B)" << endl;
  cout << flush;

  return graph;
}

void Comm::sendStack(int process, StackGraphItem **stack, int count, int edges) {
  MPI_Request request;
  const int length = edges * count * sizeof(char) + count * sizeof(int) + 2 * sizeof(int);

  if(length > MAXLENGTH) {
    cout << "Proces " << rank << ": ";
    cout << "Maximalni povolena delka zpravy je " << MAXLENGTH << "B (pozadovano " << length << "B)" << endl;
    cout << flush;

    exit(1);
  }

  char buffer[length];
  int position = 0;

  MPI_Pack(&count, 1, MPI_INT, buffer, length, &position, MPI_COMM_WORLD);
  MPI_Pack(&edges, 1, MPI_INT, buffer, length, &position, MPI_COMM_WORLD);

  for(int i = 0; i < count; i++) {
    bool *potent = stack[i]->getGraph();
    int aux = stack[i]->getEdges();
    MPI_Pack(&aux, 1, MPI_INT, buffer, length, &position, MPI_COMM_WORLD);

    for(int j = 0; j < edges; j++)
      if(potent[j]) MPI_Pack(&ONE, 1, MPI_CHAR, buffer, length, &position, MPI_COMM_WORLD);
      else MPI_Pack(&ZERO, 1, MPI_CHAR, buffer, length, &position, MPI_COMM_WORLD);

    delete [] potent;
  }
     
  MPI_Isend(buffer, position, MPI_PACKED, process, STACK, MPI_COMM_WORLD, &request);

#if DEBUG == 1
  cout << "Proces " << rank << ": ";
  cout << "Procesu " << process << " odeslano " << count << " konfiguraci (" << position << "B)" << endl;
  cout << flush;
#endif
}

void Comm::receiveStack(StackGraph *stack) {
  const int length = MAXLENGTH;
  char buffer[length];
  int position = 0;

  MPI_Recv(buffer, length, MPI_PACKED, MPI_ANY_SOURCE, STACK, MPI_COMM_WORLD, &status);

  int count = 0;
  int edges = 0;
  MPI_Unpack(buffer, length, &position, &count, 1, MPI_INT, MPI_COMM_WORLD);
  MPI_Unpack(buffer, length, &position, &edges, 1, MPI_INT, MPI_COMM_WORLD);

  for(int i = 0; i < count; i++) {
    bool *potent = new bool [edges];
    int aux = 0;
    MPI_Unpack(buffer, length, &position, &aux, 1, MPI_INT, MPI_COMM_WORLD);

    for(int j = 0; j < edges; j++)
      MPI_Unpack(buffer, length, &position, &potent[j], 1, MPI_CHAR, MPI_COMM_WORLD);

#if DEBUG == 1
    cout << "Proces " << rank << ": Prijata konfigurace ";

    for(int j = 0; j < edges; j++)
      cout << potent[j] << " ";

    cout << endl;
#endif

    stack->push(potent, aux, edges);
  }

#if DEBUG == 1
  cout << "Proces " << rank << ": ";
  cout << "Prijato " << count << " konfiguraci od procesu " << status.MPI_SOURCE << " (" << position << "B)" << endl;
  cout << flush;
#endif
}

void Comm::sendSolution(int solution) {
  for(int i = 0; i < processes; i++)
    if(i != rank) {
      MPI_Send(&solution, 1, MPI_INT, i, SOLUTION, MPI_COMM_WORLD);

      cout << "Proces " << rank << ": ";
      cout << "Procesu " << i << " odeslano reseni (" << solution << ")" << endl;
    }

  if(processes == 1) {
    cout << "Proces " << rank << ": ";
    cout << "Nalezeno reseni (" << solution << ")" << endl;
  }

  cout << flush;
}

int Comm::receiveSolution() {
  int res;

  MPI_Recv(&res, 1, MPI_INT, MPI_ANY_SOURCE, SOLUTION, MPI_COMM_WORLD, &status);

  cout << "Proces " << rank << ": ";
  cout << "Prijato reseni od procesu " << status.MPI_SOURCE << " (" << res << ")" << endl;
  cout << flush;

  return res;
}

void Comm::requestJob() {
  if(!requestActive && (processes > 1)) {
    requestActive = true;
    requestTotal++;

    MPI_Send(&ONE, 1, MPI_CHAR, donor, JOB_REQ, MPI_COMM_WORLD);

#if DEBUG == 1
    cout << "Proces " << rank << ": Odeslana zadost o praci procesu " << donor << endl;
    cout << flush;
#endif

    donor++;
    donor %= processes;

    if(donor == rank) {
      donor++;
      donor %= processes;
    }
  }
}

void Comm::ackJob() {
  char res;

  MPI_Recv(&res, 1, MPI_CHAR, MPI_ANY_SOURCE, JOB_REQ, MPI_COMM_WORLD, &status);

#if DEBUG == 1
  cout << "Proces " << rank << ": ";
  cout << "Prijata zadost o praci od procesu " << status.MPI_SOURCE << endl;
  cout << flush;
#endif
}

void Comm::receiveJob(StackGraph *stack) {
  requestActive = false;

  receiveStack(stack);
  if(!stack->empty()) requestSucc++;

#if DEBUG == 1
  if(stack->empty()) cout << "Proces " << rank << ": Zadna prace nebyla pridelena" << endl;
  else cout << "Proces " << rank << ": Prace byla pridelena" << endl;

  cout << flush;
#endif
}

void Comm::sendToken(int pProcess, int pToken) {
  if(processes > 1) {
    if(rank == 0) inquiry = true;

    MPI_Send(&pToken, 1, MPI_INT, pProcess, TOKEN, MPI_COMM_WORLD);

#if DEBUG == 1
    cout << "Proces " << rank << ": ";
    cout << "Procesu " << pProcess << " odeslan pesek (" << pToken << ")" << endl;
#endif
  }
  else {
    terminated = true;

#if DEBUG == 1
    cout << "Proces " << rank << ": ";
    cout << "Pozadano o ukonceni" << endl;
#endif
  }

#if DEBUG == 1 
  cout << flush;
#endif
}

int Comm::receiveToken() {
  int res;

  MPI_Recv(&res, 1, MPI_INT, MPI_ANY_SOURCE, TOKEN, MPI_COMM_WORLD, &status);

#if DEBUG == 1
  cout << "Proces " << rank << ": ";
  cout << "Prijat pesek od procesu " << status.MPI_SOURCE << " (" << res << ")" << endl;
  cout << flush;
#endif

  return res;
}

void Comm::sendTerminate(int pSolProcess) {
  if(rank == 0) {
    for(int i = 1; i < processes; i++) {
      if(i == pSolProcess)
	MPI_Send(&ONE, 1, MPI_CHAR, i, TERMINATE, MPI_COMM_WORLD);
      else MPI_Send(&ZERO, 1, MPI_CHAR, i, TERMINATE, MPI_COMM_WORLD);

      cout << "Proces " << rank << ": ";
      cout << "Procesu " << i << " odeslan prikaz k ukonceni" << endl;
      cout << flush;
    }

    if(rank == pSolProcess) terminated = true; // nemusim cekat na reseni
  }
}

void Comm::acceptTerminate(bool *pSolution, int pEdges) {
  char flag;

  MPI_Recv(&flag, 1, MPI_CHAR, MPI_ANY_SOURCE, TERMINATE, MPI_COMM_WORLD, &status);

  if(flag == ONE) sendConfig(status.MPI_SOURCE, pSolution, pEdges);
  terminated = true;

  cout << "Proces " << rank << ": ";
  cout << "Prijat pozadavek na ukonceni vypoctu od procesu " << status.MPI_SOURCE << endl;
  cout << flush;
}

void Comm::sendConfig(int pDest, bool *pSolution, int pEdges) {
  MPI_Request request;
  const int length = sizeof(int) + sizeof(char) * pEdges;

  if(length > MAXLENGTH) {
    cout << "Proces " << rank << ": ";
    cout << "Maximalni povolena delka zpravy je " << MAXLENGTH << "B (pozadovano " << length << "B)" << endl;
    cout << flush;

    exit(1);
  }

  char buffer[length];
  int position = 0;

  MPI_Pack(&pEdges, 1, MPI_INT, buffer, length, &position, MPI_COMM_WORLD);

  for(int i = 0; i < pEdges; i++)
    if(pSolution[i]) MPI_Pack(&ONE, 1, MPI_CHAR, buffer, length, &position, MPI_COMM_WORLD);
    else MPI_Pack(&ZERO, 1, MPI_CHAR, buffer, length, &position, MPI_COMM_WORLD);

  MPI_Isend(buffer, position, MPI_PACKED, pDest, CONFIG, MPI_COMM_WORLD, &request);

  cout << "Proces " << rank << ": ";
  cout << "Odeslana optimalni konfigurace procesu " << pDest << " (" << position << "B)" << endl;
  cout << flush;
}

void Comm::receiveConfig(bool **pSolution) {
  const int length = MAXLENGTH;
  char buffer[length];
  int position = 0;

  MPI_Recv(buffer, length, MPI_PACKED, MPI_ANY_SOURCE, CONFIG, MPI_COMM_WORLD, &status);

  int edges;
  MPI_Unpack(buffer, length, &position, &edges, 1, MPI_INT, MPI_COMM_WORLD);

  bool res[edges];

  for(int i = 0; i < edges; i++)
    MPI_Unpack(buffer, length, &position, &res[i], 1, MPI_CHAR, MPI_COMM_WORLD);

  delete [] (*pSolution);
  (*pSolution) = res;

  cout << "Proces " << rank << ": ";
  cout << "Prijata optimalni konfigurace od procesu " << status.MPI_SOURCE << " (" << position << "B)" << endl;
  cout << flush;

  terminated = true;
}

void Comm::start() {
  MPI_Barrier(MPI_COMM_WORLD);
  startTime = MPI_Wtime();

  cout << "Proces " << rank << ": ";
  cout << "Bariera (" << startTime << ", celkem " << totalTime << ")" << endl;
  cout << flush;
}

void Comm::stop() {
  MPI_Barrier(MPI_COMM_WORLD);
  endTime = MPI_Wtime();
  totalTime = endTime - startTime;

  cout << "Proces " << rank << ": ";
  cout << "Bariera (" << endTime << ", celkem " << totalTime << ")" << endl;
  cout << "(Procento uspesnych zadosti o praci je ";
  if(requestTotal > 0) cout << ((100 * requestSucc) / requestTotal);
  else cout << 0;
  cout << ", tj. " << requestSucc << " z " << requestTotal << " zadosti)" << endl;
  cout << flush;
}

