#!/bin/bash
# Submit script for PBS, Antonin Kral 2001, Martin Kacer 2002-3

# === number of processors to be used ===
#PBS -l nodes=4

# === other arguments for qsub ===
# Every line beginning with "#PBS" is used as if it was a command-line
# argument for "qsub" command.
# For example, the following line causes stout and stderr of the job
# to be intermixed and joined into a single file.
#PBS -j oe
#PBS -e /dev/null

# Other useful properties you can configure include:
# - if you want to receive email notifications at job startup/termination
# - which PBS queue to use
# - when to run the job
# - etc.
# Please see "man qsub" for the complete list of all possibilities.

# === working directory and command of your job ===
MYDIR="$HOME/mpi"
#MYPRG='./priklad1 -arg1 -arg2="value"'
MYPRG='./a.out graph'

# === which communication library to use ===
# "p4" means Ethernet, "gm" means Myrinet, "auto" tries to detect it
# Note that you need to compile your MPI program in a specific way
# if you want to use Myrinet fast switching network. You will learn how
# to do it during your seminars. Until then, please leave the following
# lines untouched.
#MYCOM="p4"
#MYCOM="gm"
MYCOM="auto"


########################################################################
# NO EDITING SHOULD BE REQUIRED BELOW THIS LINE
#
# Anyway, feel free to experiment and edit the rest of the file if you
# want to. But, do not complain to us if you break something!
########################################################################


export PATH=/bin:/usr/local/bin:/usr/bin:/nfs_exported/bin


if ! cd "$MYDIR"
then
	echo "***** Working directory does not exist: $MYDIR"
	exit 1
fi


# automatic detection of Ethernet/Myrinet library
if [ "$MYCOM" == "auto" ]
then
	PRGNAME=`echo "$MYPRG" | cut -f1 -d' '`
	if ldd "$PRGNAME" | grep -q libgm
	then
		MYCOM="gm"
	else
		MYCOM="p4"
	fi
fi


NPROC=`wc -l < $PBS_NODEFILE | tr -d " "`

echo "Current direcotory    : `pwd`"
echo "Binary                : $MYPRG"
echo "Number of processors  : $NPROC"
echo "Communication library : $MYCOM"
echo "My job ID             : $PBS_JOBID"

echo -n "Start time: "
date
echo "---------- Program output BEGINS here ----------"

mpiexec -comm="mpich-$MYCOM" $MYPRG

echo "---------- Program output ENDS here ----------"
echo -n "End time: "
date








